{
  (* File: lexer.mll

      Copyright (C) 2005-

        Jane Street Holding, LLC
        Author: Markus Mottl
        email: mmottl@janestcapital.com
        WWW: http://www.janestcapital.com/ocaml

     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
     License as published by the Free Software Foundation; either
     version 2 of the License, or (at your option) any later version.

     This library is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser General Public License for more details.

     You should have received a copy of the GNU Lesser General Public
     License along with this library; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *)

  (** Lexer: Lexer Specification for S-expressions *)

  open Printf
  open Lexing
  open Parser

  let char_for_backslash = function
    | 'n' -> '\n'
    | 't' -> '\t'
    | 'b' -> '\b'
    | 'r' -> '\r'
    | c   -> c

  let decimal_code  c d u =
    100 * (Char.code c - 48) + 10 * (Char.code d - 48) + (Char.code u - 48)

  let char_for_hexadecimal_code d u =
    let d1 = Char.code d in
    let val1 =
      if d1 >= 97 then d1 - 87
      else if d1 >= 65 then d1 - 55
      else d1 - 48 in
    let d2 = Char.code u in
    let val2 =
      if d2 >= 97 then d2 - 87
      else if d2 >= 65 then d2 - 55
      else d2 - 48 in
    Char.chr (val1 * 16 + val2)

  let found_newline lexbuf =
    let curr_p = lexbuf.lex_curr_p in
    lexbuf.lex_curr_p <-
      {
        curr_p with
        pos_lnum = curr_p.pos_lnum + 1;
        pos_bol = max 1 (curr_p.pos_cnum - 1);
      }
}

let newline = ('\010' | '\013' | "\013\010")
let space = [' ' '\009' '\012']
let whitespace = [' ' '\010' '\013' '\009' '\012']
let backslash_escapes = ['\\' '"' '\'' 'n' 't' 'b' 'r']

rule main = parse
  | newline { found_newline lexbuf; main lexbuf }
  | space+ { main lexbuf }
  | ';' [^ '\n' '\r']+ { main lexbuf }
  | '(' { LPAREN }
  | ')' { RPAREN }
  | '"'
      {
        let buf = Buffer.create 256 in
        scan_string buf lexbuf;
        STRING (Buffer.contents buf)
      }
  | ([^ ';' '(' ')' '"'] # whitespace)+ as str { STRING str }
  | eof { EOF }

and scan_string buf = parse
  | '"' { () }
  | '\\' newline [' ' '\009']* { found_newline lexbuf; scan_string buf lexbuf }
  | '\\' (backslash_escapes as c)
      {
        Buffer.add_char buf (char_for_backslash c);
        scan_string buf lexbuf
      }
  | '\\' (['0'-'9'] as c) (['0'-'9'] as d) (['0'-'9']  as u)
      {
        let v = decimal_code c d u in
        if v > 255 then (
          let pos = lexbuf.lex_curr_p in
          let msg =
            sprintf
              "Sexplib.Lexer.scan_string: \
               illegal escape at line %d char %d: `\\%c%c%c'"
              pos.pos_lnum (pos.pos_cnum - pos.pos_bol - 3)
              c d u in
          failwith msg);
        Buffer.add_char buf (Char.chr v);
        scan_string buf lexbuf
      }
  | '\\' 'x' (['0'-'9' 'a'-'f' 'A'-'F'] as d) (['0'-'9' 'a'-'f' 'A'-'F'] as u)
      {
        Buffer.add_char buf (char_for_hexadecimal_code d u);
        scan_string buf lexbuf
      }
  | '\\' (_ as c)
      {
        Buffer.add_char buf '\\';
        Buffer.add_char buf c;
        scan_string buf lexbuf
      }
    | newline as nl
        {
          found_newline lexbuf;
          Buffer.add_string buf nl;
          scan_string buf lexbuf
        }
  | [^ '\\' '"']+ as str
      {
        Buffer.add_string buf str;
        scan_string buf lexbuf
      }
  | eof { failwith "Sexplib.Lexer.scan_string: unterminated string" }
