(* File: sexp.ml

    Copyright (C) 2005-

      Jane Street Holding, LLC
      Author: Markus Mottl
      email: mmottl@janestcapital.com
      WWW: http://www.janestcapital.com/ocaml

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(* Sexp: Module for Handling S-expressions (I/O, etc.) *)

open Format

include Type

(* Input Functions *)

let input_sexp ic =
  let lexbuf = Lexing.from_channel ic in
  Parser.sexp Lexer.main lexbuf

let input_sexps ic =
  let lexbuf = Lexing.from_channel ic in
  List.rev (Parser.rev_sexps Lexer.main lexbuf)

let input_fold_sexps ~f ~init ic =
  let lexbuf = Lexing.from_channel ic in
  let rec loop acc =
    match Parser.maybe_sexp Lexer.main lexbuf with
    | None -> acc
    | Some sexp -> loop (f sexp acc) in
  loop init

let input_cnv_sexps cnv ic =
  let coll sexp acc = cnv sexp :: acc in
  List.rev (input_fold_sexps ~f:coll ~init:[] ic)


(* Escaping of Strings Used as Atoms in S-expressions *)

let is_special_char c =
  c <= ' ' || c = '"' || c = '(' || c = ')' || c = ';' || c = '\\'

let must_escape str =
  let len = String.length str in
  len = 0 ||
    let rec loop ix = is_special_char str.[ix] || ix > 0 && loop (ix - 1) in
    loop (len - 1)

let maybe_esc_str str =
  if must_escape str then
    let estr = String.escaped str in
    let elen = String.length estr in
    let res = String.create (elen + 2) in
    String.blit estr 0 res 1 elen;
    res.[0] <- '"';
    res.[elen + 1] <- '"';
    res
  else str

let pp_maybe_esc_str ppf str = pp_print_string ppf (maybe_esc_str str)


(* Output of S-expressions to I/O-channels *)

let rec pp_hum ppf = function
  | Atom str -> pp_maybe_esc_str ppf str
  | List (h :: t) ->
      pp_open_box ppf 2;
      pp_print_string ppf "(";
      pp_hum ppf h;
      pp_hum_rest ppf t
  | List [] -> pp_print_string ppf "()"

and pp_hum_rest ppf = function
  | h :: t ->
      pp_print_space ppf ();
      pp_hum ppf h;
      pp_hum_rest ppf t
  | [] ->
      pp_print_string ppf ")";
      pp_close_box ppf ()

let rec pp_mach ppf = function
  | Atom str -> pp_maybe_esc_str ppf str
  | List (h :: t) ->
      pp_print_string ppf "(";
      pp_mach ppf h;
      pp_mach_rest ppf t
  | List [] -> pp_print_string ppf "()"

and pp_mach_rest ppf = function
  | h :: t ->
      pp_print_string ppf " ";
      pp_mach ppf h;
      pp_mach_rest ppf t
  | [] -> pp_print_string ppf ")"

let pp = pp_mach


(* Output of S-expressions to Formatters *)

let output_hum oc sexp =
  let ppf = formatter_of_out_channel oc in
  pp_hum ppf sexp;
  pp_print_flush ppf ()

let rec output_mach oc = function
  | Atom str -> output_string oc (maybe_esc_str str)
  | List (h :: t) ->
      output_char oc '(';
      output_mach oc h;
      output_mach_loop_rest oc t
  | List [] -> output_string oc "()"

and output_mach_loop_rest oc = function
  | h :: t -> output_char oc ' '; output_mach oc h; output_mach_loop_rest oc t
  | [] -> output_char oc ')'

let output = output_mach


(* String Conversions *)

let of_string str =
  let lexbuf = Lexing.from_string str in
  Parser.sexp Lexer.main lexbuf

let to_string_hum sexp =
  let buf = Buffer.create 1024 in
  Format.bprintf buf "%a" pp_hum sexp;
  Buffer.contents buf

let to_string_mach sexp =
  let buf = Buffer.create 1024 in
  let rec loop = function
    | Atom str -> Buffer.add_string buf (maybe_esc_str str)
    | List (h :: t) -> Buffer.add_char buf '('; loop h; loop_rest t
    | List [] -> Buffer.add_string buf "()"
  and loop_rest = function
    | h :: t -> Buffer.add_char buf ' '; loop h; loop_rest t
    | [] -> Buffer.add_char buf ')' in
  loop sexp;
  Buffer.contents buf

let to_string = to_string_mach


(* Utilities for Automated Type Conversions *)

let unit = List []

external sexp_of_t : t -> t = "%identity"
external t_of_sexp : t -> t = "%identity"
