(* File: sexp.mli

    Copyright (C) 2005-

      Jane Street Holding, LLC
      Author: Markus Mottl
      email: mmottl@janestcapital.com
      WWW: http://www.janestcapital.com/ocaml

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(** Sexp: Module for Handling S-expressions (I/O, etc.) *)

open Format

(** Type of S-expressions *)
type t = Type.t = Atom of string | List of t list


(** {6 Input Functions} *)

val input_sexp : in_channel -> t
(** [input_sexp ic] reads an S-expression from input channel [ic]. *)

val input_sexps : in_channel -> t list
(** [input_sexps ic] reads a list of S-expressions from input channel
    [ic]. *)

val input_fold_sexps : f:(t -> 'a -> 'a) -> init:'a -> in_channel -> 'a
(** [input_fold_sexps ~f ~init ic] folds over all S-expressions provided
    by input channel [ic] using functions [f] and initial state [init]. *)

val input_cnv_sexps : (t -> 'a) -> in_channel -> 'a list
(** [input_cnv_sexps f ic] maps all S-expressions provided by input
    channel [ic] to some list using function [f]. *)


(** {6 Output of S-expressions to I/O-channels} *)

val output_hum : out_channel -> t -> unit
(** [output_hum oc sexp] outputs S-expression [sexp] to output channel
    [oc] in human readable form. *)

val output_mach : out_channel -> t -> unit
(** [output_mach oc sexp] outputs S-expression [sexp] to output channel
    [oc] in machine readable (i.e. most compact) form. *)

val output : out_channel -> t -> unit
(** [output oc sexp] same as [output_mach]. *)


(** {6 Output of S-expressions to Formatters} *)

val pp_hum : formatter -> t -> unit
(** [pp_hum ppf sexp] outputs S-expression [sexp] to formatter [ppf]
    in human readable form. *)

val pp_mach : formatter -> t -> unit
(** [pp_mach ppf sexp] outputs S-expression [sexp] to formatter [ppf]
    in machine readable (i.e. most compact) form. *)

val pp : formatter -> t -> unit
(** [pp ppf sexp] same as [pp_mach]. *)


(** {6 String Conversions} *)

val of_string : string -> t
(** [of_string str] converts string [str] to an S-expression. *)

val to_string_hum : t -> string
(** [to_string_hum sexp] converts S-expression [sexp] to a string in
    human readable form. *)

val to_string_mach : t -> string
(** [to_string_mach sexp] converts S-expression [sexp] to a string in
    machine readable (i.e. most compact) form. *)

val to_string : t -> string
(** [to_string sexp] same as [to_string_mach]. *)


(** {6 Utilities for Automated Type Conversions} *)

val unit : t
(** [unit] the unit-value as expressed by an S-expression. *)

external sexp_of_t : t -> t = "%identity"
(** [sexp_of_t sexp] maps S-expressions which are part of a type with
    automated S-expression conversion to themselves. *)

external t_of_sexp : t -> t = "%identity"
(** [t_of_sexp sexp] maps S-expressions which are part of a type with
    automated S-expression conversion to themselves. *)
